﻿namespace MefContrib.Models.Provider
{
    using System;
    using System.Diagnostics;
    using System.Reflection;

    /// <summary>
    /// Provides an abstraction for working with a field on a reflection level.
    /// </summary>
    public class ComposableField : ComposableMember
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="ComposableField"/> class
        /// for the specified <see cref="MemberInfo"/>.
        /// </summary>
        /// <param name="member">The member to make composable.</param>
        /// <exception cref="InvalidOperationException">The specified value for the member parameter was not a <see cref="FieldInfo"/> instance.</exception>
        public ComposableField(MemberInfo member)
            : base(member)
        {
            FieldInfo info = member as FieldInfo;
            if (info == null)
            {
                throw new InvalidOperationException("The specified value for the member parameter was not a FieldInfo instance.");
            }

            this.Field = info;

            this.ValueGetter =
                getInstance =>
                this.Field.GetValue(getInstance);

            this.ValueSetter =
                (setInstance, setValue) =>
                this.Field.SetValue(setInstance, setValue);
        }

        /// <summary>
        /// Gets the reference to the field.
        /// </summary>
        /// <value>A <see cref="FieldInfo"/> object.</value>
        public FieldInfo Field { get; protected set; }

        /// <summary>
        /// Gets a value indicating whether or not type of the <see cref="ComposableField"/> is of a lazy type or not.
        /// </summary>
        /// <value><see langword="true" /> if it is of a known lazy type; otherwise <see langword="false" />.</value>
        public override bool IsLazyType
        {
            get { return this.ReturnType.IsLazyType(); }
        }

        /// <summary>
        /// Gets a value indicating whether or not a value can be retrieved from the member.
        /// </summary>
        /// <value><see langword="true" /> if a value can be retrieved; otherwise <see langword="false" />.</value>
        public override bool IsReadable
        {
            [DebuggerStepThrough]
            get { return true; }
        }

        /// <summary>
        /// Gets a value indicating whether or not an instance is needed for the value to be retrieved or assigned.
        /// </summary>
        /// <value><see langword="true" /> if an instance is needed; otherwise <see langword="false" />.</value>
        public override bool IsInstanceNeeded
        {
            [DebuggerStepThrough]           
            get { return !this.Field.IsStatic; }
        }

        /// <summary>
        /// Gets a value indicating whether or not a value can be assigned to the member.
        /// </summary>
        /// <value><see langword="true" /> if a value can be assigned; otherwise <see langword="false" />.</value>
        public override bool IsWritable
        {
            [DebuggerStepThrough]
            get { return !this.Field.IsInitOnly && !this.Field.IsLiteral; }
        }

        /// <summary>
        /// Gets the <see cref="Type"/> of the return value of the <see cref="ComposableField"/>.
        /// </summary>
        /// <value>A <see cref="Type"/> object.</value>
        public override Type ReturnType
        {
            [DebuggerStepThrough]
            get { return this.Field.FieldType; }
        }
    }
}